from typing import (
    List,
    Optional,
    Tuple,
)
from typing_extensions import (
    Final,
)
import spidev

from . import SpinDevice


class SpinChain:
    """Class for constructing a chain of SPIN devices"""
    def __init__(
            self, busy_pin: int, total_devices: int,
            spi_select: Tuple[int, int],
            chip_select_pin: Optional[int] = None) -> None:
        """
        :chip_select_pin: Chip select pin,
        if different from hardware SPI CS pin
        :busy_pin: Pin to read busy status from
        :total_devices: Total number of devices in chain
        :spi_select: A SPI bus, device pair, e.g. (0, 0)
        """
        assert(total_devices > 0)

        self._chip_select_pin: Final   = chip_select_pin
        self._busy_pin: Final          = busy_pin
        self._total_devices: Final     = total_devices

        # {{{ SPI setup
        self._spi: Final = spidev.SpiDev()

        bus, device = spi_select
        self._spi.open(bus, device)

        self._spi.mode = 3
        # Device expects MSB to be sent first
        self._spi.lsbfirst = False
        self._spi.max_speed_hz = 1000000
        # CS pin is active low
        self._spi.cshigh = False
        # }}}

    def create(self, position: int) -> SpinDevice:
        """
                   +----------+
              MOSI |   MCU    | MISO
       +-----------+          +---------------+
       |           +----------+               |
       |                                      |
       |                                      |
       |             SPIN ICs                 |
       |   +-----+     +-----+     +-----+    |
       |SDI|     |     |     |     |     |SDO |
       +---+  2  +-----+  1  +-----+  0  +----+
           |     |     |     |     |     |
           |     |     |     |     |     |
           +-----+     +-----+     +-----+
        Create a new SPIN device at the specified chain location
        :position: Device position in chain
        :return: A newly-instantiated SpinDevice

        """
        assert(position >= 0)
        assert(position < self._total_devices)

        return SpinDevice(
            position,
            self._busy_pin,
            self._total_devices,
            self._spi,
            self._chip_select_pin,
        )



from typing_extensions import (
    Final,
)


class Status:
    HiZ: Final                  = 0x0001
    NotBusy: Final              = 0x0002  # active low
    SwitchFlag: Final           = 0x0004  # low on closed switch, high on open
    SwitchEvent: Final          = 0x0008  # high on falling edge
    Dir: Final                  = 0x0010
    CmdNotPerformed: Final      = 0x0080
    CmdWrong: Final             = 0x0100
    NotUndervoltage: Final      = 0x0200  # active low
    NotThermalWarning: Final    = 0x0400  # active low
    NotThermalShutdown: Final   = 0x0800  # active low
    NotOvercurrent: Final       = 0x1000  # active low
    NotStepLossA: Final         = 0x2000  # low on stall detect
    NotStepLossB: Final         = 0x4000  # low on stall detect
    StepClockMode: Final        = 0x8000


class MotorStatus:
    Offset: Final           = 5
    Accelerating: Final     = 0b01 << Offset
    Decelerating: Final     = 0b10 << Offset
    ConstantSpeed: Final    = 0b11 << Offset
    Stopped: Final          = 0b00 << Offset


class SwitchStatus:
    EventNone: Final        = 0
    EventFallingEdge: Final = Status.SwitchEvent  # Switch just closed

    Closed: Final           = 0
    Open: Final             = Status.SwitchFlag



from typing import (
    Dict,
)
from typing_extensions import (
    Final,
)


class Command:
    GoHome: Final           = 0x70
    GoMark: Final           = 0x78
    GoTo: Final             = 0x60
    GoToDir: Final          = 0x68  # ORed with DIR
    GoUntil: Final          = 0x82  # ORed with ACT, DIR
    HiZHard: Final          = 0xA8
    HiZSoft: Final          = 0xA0
    Nop: Final              = 0x00
    Move: Final             = 0x40  # ORed with DIR. Unuseable while running
    ParamGet: Final         = 0x20  # ORed with target register value
    ParamSet: Final         = 0x00  # ORed with target register value
    ReleaseSw: Final        = 0x92  # ORed with ACT, DIR
    ResetDevice: Final      = 0xC0
    ResetPos: Final         = 0xD8  # Clears ABS_POS
    Run: Final              = 0x50  # ORed with DIR
    StatusGet: Final        = 0xD0
    StepClock: Final        = 0x58  # ORed with DIR
    StopHard:  Final        = 0xB8
    StopSoft: Final         = 0xB0

    @staticmethod
    def getPayloadSize(command: int) -> int:
        """return size of payload in bytes expected to follow command

        :command: command to query
        :returns: payload size in bytes

        """
        assert(command in PayloadSize)

        return PayloadSize[command]


PayloadSize: Final[Dict[int, int]] = {
    Command.GoHome:         0,
    Command.GoMark:         0,
    Command.GoTo:           3,
    Command.GoToDir:        3,
    Command.GoUntil:        3,
    Command.HiZHard:        0,
    Command.HiZSoft:        0,
    Command.Nop:            0,
    Command.Move:           3,
    Command.ReleaseSw:      0,
    Command.ResetDevice:    0,
    Command.ResetPos:       0,
    Command.Run:            3,
    Command.StepClock:      0,
    Command.StopHard:       0,
    Command.StopSoft:       0,
}



from typing import (
    List,
)


def getByteCount(value: int) -> int:
    """Calculate the number of bytes required to represent value

    :value: value to check. Non-negative
    :returns: number of bytes to represent the number

    """

    assert(value >= 0)

    if 0 == value:
        return 1

    return (value.bit_length() + 7) // 8


def resizeToLength(array: List[int], length: int) -> List[int]:
    """Resizes array, 0-extending the first positions,
    or truncating the first positions

    :array: Array to resize (if at all)
    :length: Desired length of array
    :returns: Resized array or original array

    """
    assert(length >= 0)
    difference = abs(len(array) - length)

    if len(array) > length:
        return array[difference:]

    return ([0] * difference) + array


def toByteArray(value: int) -> List[int]:
    """Splits an integer into a list of bytes

    :value: Value to convert. Must be non-negative
    :returns: List of bytes, with MSB at entry 0

    """
    byte_count = getByteCount(value)
    return list(value.to_bytes(byte_count, byteorder='big'))


def toByteArrayWithLength(value: int, length: int) -> List[int]:
    """Splits an integer into a list of bytes
    First bytes will be truncated or padded with 0 as
    required by length

    :value: Value to convert. Must be non-negative
    :length: Desired length in bytes
    :returns: List of bytes, with MSB at entry 0

    """
    return resizeToLength(toByteArray(value), length)


def toInt(byte_array: List[int]) -> int:
    """Convert a byte array to an integer

    :byte_array: Byte array with MSB first
    :returns: Single integer equivalent to given byte array

    """
    multiplier = 256 ** len(byte_array)
    result = 0

    for byte in byte_array:
        multiplier //= 256
        result += byte * multiplier

    return result



from typing import (
    Dict,
)
from typing_extensions import (
    Final,
)


class Register:
    Acc: Final              = 0x05
    AdcOut: Final           = 0x12
    AlarmEn: Final          = 0x17
    Dec: Final              = 0x06
    Config: Final           = 0x18
    KTherm: Final           = 0x11
    KvalAcc: Final          = 0x0B
    KvalDec: Final          = 0x0C
    KvalHold: Final         = 0x09
    KvalRun: Final          = 0x0A
    Mark: Final             = 0x03
    PosAbs: Final           = 0x01
    PosEl: Final            = 0x02
    SlpFnAcc: Final         = 0x0F
    SlpFnDec: Final         = 0x10
    SlpSt: Final            = 0x0E
    Speed: Final            = 0x04
    SpeedFS: Final          = 0x15
    SpeedInt: Final         = 0x0D
    SpeedMax: Final         = 0x07
    SpeedMin: Final         = 0x08
    Status: Final           = 0x19
    StepMode: Final         = 0x16
    ThOcd: Final            = 0x13
    ThStl: Final            = 0x14

    @staticmethod
    def getSize(register: int) -> int:
        """get the register size in bytes

        :register: Register to check
        :returns: Register size in bytes

        """
        assert(register in RegisterSize)

        return RegisterSize[register]


RegisterSize: Final[Dict[int, int]] = {
    Register.Acc:       2,
    Register.AdcOut:    1,
    Register.AlarmEn:   1,
    Register.Dec:       2,
    Register.Config:    2,
    Register.KTherm:    1,
    Register.KvalAcc:   1,
    Register.KvalDec:   1,
    Register.KvalHold:  1,
    Register.KvalRun:   1,
    Register.Mark:      3,
    Register.PosAbs:    3,
    Register.PosEl:     2,
    Register.SlpFnAcc:  1,
    Register.SlpFnDec:  1,
    Register.SlpSt:     1,
    Register.Speed:     3,
    Register.SpeedFS:   2,
    Register.SpeedInt:  2,
    Register.SpeedMax:  2,
    Register.SpeedMin:  2,
    Register.Status:    2,
    Register.StepMode:  1,
    Register.ThOcd:     1,
    Register.ThStl:     1,
}



from typing import (
    List,
    Optional,
    Tuple,
)
from typing_extensions import (
    Final,
)

from .constants import (
    Command,
    Constant,
    Register,
    Status,
)
from .utility import (
    toByteArrayWithLength,
    toInt,
)


class SpiStub:
    def xfer2(self, data: List[int]) -> List[int]:
        pass


class SpinDevice:
    """Class providing access to a single SPIN device"""

    def __init__(
            self, position: int, busy_pin: int,
            total_devices: int, spi: SpiStub,
            chip_select_pin: Optional[int] = None):
        """
        :position: Position in chain, where 0 is the last device in chain
        :chip_select_pin: Chip select pin,
        if different from hardware SPI CS pin
        :busy_pin: Pin to read busy status from
        :total_devices: Total number of devices in chain
        :spi: SPI object used for serial communication
        """
        self._position: Final           = position
        self._chip_select_pin: Final    = chip_select_pin
        self._busy_pin: Final           = busy_pin
        self._total_devices: Final      = total_devices
        self._spi: Final                = spi

        self._direction                 = Constant.DirForward

    def _write(self, data: int) -> int:
        """Write a single byte to the device.

        :data: A single byte representing a command or value
        :return: Returns response byte
        """
        assert(data >= 0)
        assert(data <= 0xFF)

        buffer = [0] * self._total_devices
        buffer[self._position] = data

        # TODO: CS LOW
        response = self._spi.xfer2(buffer)
        # TODO: CS HIGH

        return response[self._position]

    def _writeMultiple(self, data: List[int]) -> int:
        """Write each byte in list to device
        Used to combine calls to _write

        :data: List of single byte values to send
        :return: Response bytes as int
        """
        response = []

        for data_byte in data:
            response.append(self._write(data_byte))

        return toInt(response)

    def _writeCommand(
            self, command: int,
            payload: Optional[int] = None,
            payload_size: Optional[int] = None) -> int:
        """Write command to device with payload (if any)

        :command: Command to write
        :payload: Payload (if any)
        :payload_size: Payload size in bytes
        :return: Response bytes as int
        """
        response = self._write(command)

        if payload is None:
            return response

        return self._writeMultiple(
            toByteArrayWithLength(payload, payload_size)
        )

    def setRegister(self, register: int, value: int) -> None:
        """Set the specified register to the given value
        :register: The register location
        :value: Value register should be set to
        """
        RegisterSize = Register.getSize(register)
        set_command = Command.ParamSet | register

        self._writeCommand(set_command, value, RegisterSize)

    def getRegister(self, register: int) -> int:
        """Fetches a register's contents and returns the current value

        :register: Register location to be accessed
        :returns: Value of specified register

        """
        RegisterSize = Register.getSize(register)
        self._writeCommand(Command.ParamGet | register)

        return self._writeMultiple([Command.Nop] * RegisterSize)

    def move(self, steps: int) -> None:
        """Move motor n steps

        :steps: Number of (micro)steps to take

        """
        assert(steps >= 0)
        assert(steps <= Constant.MaxSteps)

        PayloadSize = Command.getPayloadSize(Command.Move)

        self._writeCommand(Command.Move | self._direction, steps, PayloadSize)

    def run(self, steps_per_second: float) -> None:
        """Run the motor at the given steps per second

        :steps_per_second: Full steps per second up to 15625.
        0.015 step/s resolution

        """
        assert(steps_per_second >= 0)
        assert(steps_per_second <= Constant.MaxStepsPerSecond)

        speed = int(steps_per_second * Constant.SpsToSpeed)
        PayloadSize = Command.getPayloadSize(Command.Run)

        self._writeCommand(Command.Run | self._direction, speed, PayloadSize)

    def setDirection(self, direction: int) -> None:
        """Set motor direction. Does not affect active movement

        :direction: Direction as declared in Constant

        """
        assert(direction >= 0)
        assert(direction < Constant.DirMax)

        self._direction = direction

    def hiZHard(self) -> None:
        """Stop motors abruptly, release holding current

        """
        self._writeCommand(Command.HiZHard)

    def hiZSoft(self) -> None:
        """Stop motors, release holding current

        """
        self._writeCommand(Command.HiZSoft)

    def stopHard(self) -> None:
        """Stop motors abruptly, maintain holding current

        """
        self._writeCommand(Command.StopHard)

    def stopSoft(self) -> None:
        """Stop motors, maintain holding current

        """
        self._writeCommand(Command.StopSoft)

    def getStatus(self) -> int:
        """Get status register
        Resets alarm flags. Does not reset HiZ
        :returns: 2 bytes status as an int

        """
        return self._writeCommand(Command.Status)

    def isBusy(self) -> bool:
        """Checks busy status of the device
        :returns: True if device is busy, else False

        """
        # We use getRegister instead of getStatus
        # So as not to clear any warning flags
        status = self.getRegister(Register.Status)

        return False if (status & Status.NotBusy) else True



from typing_extensions import (
    Final,
)


class Constant:
    DirReverse: Final           = 0
    DirForward: Final           = 1
    DirMax: Final               = 2

    ActResetPos: Final          = (0 << 3)
    ActSetMark: Final           = (1 << 3)

    MaxStepsPerSecond: Final[float]     = 15625.0
    MaxSteps: Final                     = int(2 ** 22)

    TickSeconds: Final[float]           = 250 * (10 ** -9)
    SpsToSpeed: Final[float]            = TickSeconds / (2 ** -28)



from setuptools import setup, find_packages

setup(
      name='st_spin',
      version='0.0.12',
      url='https://github.com/m-laniakea/st_spin',
      license='GPLv1',
      author='eir',
      author_email='m-laniakea@users.noreply.github.com',
      description='Interface for ST SpinFamily motor drivers',
      packages=find_packages(exclude=['tests']),
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      zip_safe=False)
